# Generated by Django 2.0.6 on 2018-06-12 15:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('lactolyse', '0002_add_results'),
    ]

    operations = [
        migrations.AddField(
            model_name='athlete',
            name='contributor',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name='lactatethresholdanalyses',
            name='contributor',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Generated by Django 2.0.6 on 2018-06-11 21:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [('lactolyse', '0001_initial')]

    operations = [
        migrations.AddField(
            model_name='lactatethresholdanalyses',
            name='result_at2',
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name='lactatethresholdanalyses',
            name='result_at4',
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name='lactatethresholdanalyses',
            name='result_cross',
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name='lactatethresholdanalyses',
            name='result_dmax',
            field=models.IntegerField(null=True),
        ),
    ]



"""Views for Lactate Threshold Analysis."""
import os

from asgiref.sync import async_to_sync

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db import transaction
from django.forms import formset_factory
from django.urls import reverse_lazy

from channels.layers import get_channel_layer

from lactolyse.forms import AthleteForm, ConconiMeasurementForm
from lactolyse.models import ConconiTestAnalyses
from lactolyse.protocol import CONCONI_REPORT_TYPE, GROUP_SESSIONS, RUN_ANALYSIS_CHANNEL

from .base import MultiFormView


class ConconiTestView(LoginRequiredMixin, MultiFormView):
    """Conconi Test view."""

    template_name = os.path.join('lactolyse', 'conconi_test.html')
    analyses_name = 'conconi_test'
    success_url = reverse_lazy('analyses_success')

    form_classes = {
        'athlete': AthleteForm,
        'measurements': formset_factory(
            ConconiMeasurementForm, min_num=7, extra=20, validate_min=True
        ),
    }

    def get_initial(self):
        """Return initial values if ``DEBUG`` setting is set to ``True``."""
        if not settings.DEBUG:
            return {}

        return {
            'athlete': {'name': "Domen Blenkuš", 'age': "28", 'weight': 74.5},
            'measurements': [
                {'power': 120, 'heart_rate': 128},
                {'power': 140, 'heart_rate': 135},
                {'power': 160, 'heart_rate': 145},
                {'power': 180, 'heart_rate': 143},
                {'power': 200, 'heart_rate': 147},
                {'power': 220, 'heart_rate': 157},
                {'power': 240, 'heart_rate': 159},
                {'power': 260, 'heart_rate': 163},
                {'power': 280, 'heart_rate': 167},
                {'power': 300, 'heart_rate': 170},
                {'power': 320, 'heart_rate': 172},
                {'power': 340, 'heart_rate': 175},
                {'power': 360, 'heart_rate': 177},
                {'power': 380, 'heart_rate': 180},
            ],
        }

    def _save_data(self, forms):
        """Save data."""
        with transaction.atomic():
            athlete = forms['athlete'].save(contributor=self.request.user)

            analyses = ConconiTestAnalyses.objects.create(
                contributor=self.request.user, athlete=athlete
            )

            measurements = []
            for form in forms['measurements']:
                if form.cleaned_data:
                    measurement = form.save(analyses=analyses)
                    measurements.append(measurement)

        return analyses

    def forms_valid(self, forms):
        """Save data, make report and reference it is Django's session for later views."""
        channel_layer = get_channel_layer()

        analysis = self._save_data(forms)

        session_id = self.request.session.session_key

        async_to_sync(channel_layer.send)(
            RUN_ANALYSIS_CHANNEL,
            {
                'type': CONCONI_REPORT_TYPE,
                'analysis_pk': analysis.pk,
                'notify_channel': GROUP_SESSIONS.format(websocket_id=session_id),
            },
        )

        return super().forms_valid(forms)



# Generated by Django 2.0.2 on 2018-03-26 21:05

from django.db import migrations, models
import django.db.models.deletion
import lactolyse.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Athlete',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=100)),
                ('age', models.IntegerField()),
                ('weight', models.DecimalField(decimal_places=2, max_digits=5)),
            ],
        ),
        migrations.CreateModel(
            name='LactateMeasurement',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('power', models.IntegerField()),
                ('heart_rate', models.IntegerField()),
                ('lactate', models.DecimalField(decimal_places=2, max_digits=4)),
            ],
        ),
        migrations.CreateModel(
            name='LactateThresholdAnalyses',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('created', models.DateField(auto_now_add=True)),
                (
                    'report',
                    models.FileField(
                        upload_to=lactolyse.models.generate_lactate_threshold_name
                    ),
                ),
                (
                    'athlete',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='lactolyse.Athlete',
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name='lactatemeasurement',
            name='analyses',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='lactolyse.LactateThresholdAnalyses',
            ),
        ),
    ]



"""Lactolyse base analysis."""
import os

import jinja2

from lactolyse.registry import registry

TEX_JOB_NAME = 'report'

latex_jinja_env = jinja2.Environment(
    block_start_string=r'\BLOCK{',
    block_end_string=r'}',
    variable_start_string=r'\VAR{',
    variable_end_string=r'}',
    comment_start_string=r'\#{',
    comment_end_string=r'}',
    line_statement_prefix=r'%%',
    line_comment_prefix=r'%#',
    trim_blocks=True,
    autoescape=False,
    loader=jinja2.PackageLoader(
        'lactolyse', os.path.join('templates', 'lactolyse', 'latex')
    ),
)


def validate_analysis(analysis):
    cls_name = analysis.__name__

    assert analysis.name, "Subclass '{}' must have defined 'name' attribute.".format(
        cls_name
    )
    assert (
        analysis.template
    ), "Subclass '{}' must have defined 'template' attribute.".format(cls_name)


class BaseAnalysisMeta(type):
    def __new__(mcs, name, bases, namespace):
        cls = type.__new__(mcs, name, bases, namespace)
        if not (
            cls.__module__ == 'lactolyse.analyses.base'
            and cls.__name__ == 'BaseAnalysis'
        ):
            validate_analysis(cls)
            registry.add(cls)
        return cls


class BaseAnalysis(metaclass=BaseAnalysisMeta):
    """Base class for performing analyses."""

    def __init__(self, runtime_dir):
        """Initialize analysis and set variables based on given runtime dir."""
        self._runtime_dir = runtime_dir

        self._tex_file = os.path.join(self._runtime_dir, TEX_JOB_NAME + '.tex')
        self._pdf_file = os.path.join(self._runtime_dir, TEX_JOB_NAME + '.pdf')

        self._context = None
        self._rendered_template = None

    @property
    def name(self):
        """Name of the analysis."""
        raise NotImplementedError(
            "Subclass od 'BaseAnalysis' must define 'name' attribute."
        )

    @property
    def template(self):
        """Filename of the template file."""
        raise NotImplementedError(
            "Subclass od 'BaseAnalysis' must define 'template' attribute."
        )

    def get_pdf_file(self):
        """Return the path to the PDF file."""
        return self._pdf_file

    def render_context(self, inputs):
        """Calculate context for ``render_template`` method.

        This function should be overriden in subclass.
        """
        return {}

    def run_render_context(self, *args, **kwargs):
        """Run ``render_context`` and save the result.

        This is a utility function that simplify writing of
        ``render_context`` function and remove the need of directly
        accessing ``self._context`` argument.
        """
        self._context = self.render_context(*args, **kwargs)

    @property
    def run_command(self):
        """Command for preparing report."""
        return 'xelatex -interaction=nonstopmode -jobname={0} {0}.tex'.format(
            TEX_JOB_NAME
        )

    def render_template(self):
        """Redner template and save it to a file."""
        template = latex_jinja_env.get_template(self.template)
        self._rendered_template = template.render(**self._context)

        with open(self._tex_file, 'w', encoding="utf-8") as fn:
            fn.write(self._rendered_template)

    def get_results(self, context):
        """Return the result of the analysis.

        This function should be overriden in subclass.
        """
        return {}

    def run_get_results(self):
        """Run ``get_results`` and pass the context to it.

        This is a utility function that simplify writing of
        ``get_results`` function and remove the need of directly
        accessing ``self._context`` argument.
        """
        return self.get_results(self._context)



from django.conf.urls import url

from channels.routing import ChannelNameRouter, ProtocolTypeRouter, URLRouter
from channels.sessions import SessionMiddlewareStack

from lactolyse.consumers import ClientConsumer, RunAnalysisConsumer
from lactolyse.protocol import RUN_ANALYSIS_CHANNEL


application = ProtocolTypeRouter(
    {
        # Client-facing consumers.
        'websocket': SessionMiddlewareStack(URLRouter([url(r'^ws/$', ClientConsumer)])),
        # Background worker consumers.
        'channel': ChannelNameRouter({RUN_ANALYSIS_CHANNEL: RunAnalysisConsumer}),
    }
)



"""Lactolyse base forms."""
from django.http import HttpResponseRedirect
from django.views.generic import FormView


class MultiFormView(FormView):
    """Multiform view."""

    form_classes = {}

    def validate_forms(self, forms):
        """Check if forms are valid."""
        return all([form.is_valid() for form in forms.values()])

    def forms_valid(self, forms):
        """If the form is valid, redirect to the supplied URL."""
        return HttpResponseRedirect(self.get_success_url())

    def forms_invalid(self, forms):
        """If the form is invalid, render the invalid form."""
        return self.render_to_response(self.get_context_data(forms=forms))

    def get_forms_kwargs(self):
        """Return the keyword arguments for instantiating forms."""
        initial = self.get_initial()

        kwargs = {}
        for key in self.form_classes:
            kwargs[key] = {'initial': initial.get(key), 'prefix': self.get_prefix()}
            if self.request.method in ('POST', 'PUT'):
                kwargs[key].update(
                    {'data': self.request.POST, 'files': self.request.FILES}
                )

        return kwargs

    def get_forms(self):
        """Return instances of forms to be used in this view."""
        forms = {}
        form_kwargs = self.get_forms_kwargs()
        for key, form_class in self.form_classes.items():
            forms[key] = form_class(**form_kwargs[key])
        return forms

    def get_context_data(self, **kwargs):
        """Insert the form into the context dict."""
        kwargs.setdefault('view', self)

        if 'forms' not in kwargs:
            kwargs['forms'] = self.get_forms()

        if self.extra_context is not None:
            kwargs.update(self.extra_context)

        return kwargs

    def post(self, request, **kwargs):
        """Check the form and return the apropriate response."""
        forms = self.get_forms()

        if self.validate_forms(forms):
            return self.forms_valid(forms)
        else:
            return self.forms_invalid(forms)



"""Views for Lactate Threshold Analysis."""
import os

from asgiref.sync import async_to_sync

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.db import transaction
from django.forms import formset_factory
from django.urls import reverse_lazy

from channels.layers import get_channel_layer

from lactolyse.forms import AthleteForm, LactateMeasurementForm
from lactolyse.models import LactateThresholdAnalyses
from lactolyse.protocol import GROUP_SESSIONS, LACTATE_REPORT_TYPE, RUN_ANALYSIS_CHANNEL

from .base import MultiFormView


class LactateThresholdView(LoginRequiredMixin, MultiFormView):
    """Lactate Threshold Analisys view."""

    template_name = os.path.join('lactolyse', 'lactate_threshold.html')
    analyses_name = 'lactate_threshold'
    success_url = reverse_lazy('analyses_success')

    form_classes = {
        'athlete': AthleteForm,
        'measurements': formset_factory(
            LactateMeasurementForm, min_num=5, extra=10, validate_min=True
        ),
    }

    def get_initial(self):
        """Return initial values if ``DEBUG`` setting is set to ``True``."""
        if not settings.DEBUG:
            return {}

        return {
            'athlete': {'name': "Domen Blenkuš", 'age': "28", 'weight': 79.5},
            'measurements': [
                {'power': 80, 'heart_rate': 125, 'lactate': 2.4},
                {'power': 120, 'heart_rate': 135, 'lactate': 2.5},
                {'power': 160, 'heart_rate': 147, 'lactate': 1.6},
                {'power': 200, 'heart_rate': 160, 'lactate': 2.2},
                {'power': 240, 'heart_rate': 177, 'lactate': 4.1},
                {'power': 280, 'heart_rate': 195, 'lactate': 12.4},
            ],
        }

    def _save_data(self, forms):
        """Save data."""
        with transaction.atomic():
            athlete = forms['athlete'].save(contributor=self.request.user)

            analyses = LactateThresholdAnalyses.objects.create(
                contributor=self.request.user, athlete=athlete
            )

            measurements = []
            for form in forms['measurements']:
                if form.cleaned_data:
                    measurement = form.save(analyses=analyses)
                    measurements.append(measurement)

        return analyses

    def forms_valid(self, forms):
        """Save data, make report and reference it is Django's session for later views."""
        channel_layer = get_channel_layer()

        analysis = self._save_data(forms)

        session_id = self.request.session.session_key
        async_to_sync(channel_layer.send)(
            RUN_ANALYSIS_CHANNEL,
            {
                'type': LACTATE_REPORT_TYPE,
                'analysis_pk': analysis.pk,
                'notify_channel': GROUP_SESSIONS.format(websocket_id=session_id),
            },
        )

        return super().forms_valid(forms)



"""Django channels consumers."""
import os
import tempfile
from uuid import uuid4

from asgiref.sync import async_to_sync

from django.core.files import File
from django.urls import reverse

from channels.consumer import SyncConsumer
from channels.generic.websocket import JsonWebsocketConsumer

from lactolyse.executors import executor
from lactolyse.models import ConconiTestAnalyses, LactateThresholdAnalyses
from lactolyse.protocol import GROUP_SESSIONS
from lactolyse.utils import serialize_model_instance


class RunAnalysisConsumer(SyncConsumer):
    """Consumer for running the executor."""

    def lactolyse_lactate_report(self, event):
        """Make report for Lactate Thresold Analysis."""
        analysis = (
            LactateThresholdAnalyses.objects.select_related('athlete')
            .prefetch_related('lactatemeasurement_set')
            .get(pk=event['analysis_pk'])
        )
        measurements = analysis.lactatemeasurement_set.order_by('power').values(
            'power', 'heart_rate', 'lactate'
        )

        inputs = {
            'power': [m['power'] for m in measurements],
            'heart_rate': [m['heart_rate'] for m in measurements],
            'lactate': [float(m['lactate']) for m in measurements],
            'weight': float(analysis.athlete.weight),
        }

        report_dir = tempfile.TemporaryDirectory()
        report_filename = '{}.pdf'.format(uuid4().hex)
        report_path = os.path.join(report_dir.name, report_filename)

        results = executor.run('lactate_threshold', report_path, inputs)

        analysis.result_dmax = results['dmax']
        analysis.result_cross = results['cross']
        analysis.result_at2 = results['at2']
        analysis.result_at4 = results['at4']

        with open(report_path, 'rb') as fn:
            report_file = File(fn)
            analysis.report = report_file

            analysis.save()

        async_to_sync(self.channel_layer.group_send)(
            event['notify_channel'],
            {
                'type': 'websocket_send',
                'download_url': reverse(
                    'download_report',
                    kwargs={'ref': serialize_model_instance(analysis)},
                ),
            },
        )

    def lactolyse_conconi_report(self, event):
        """Make report for Conconi Test Analysis."""
        analysis = (
            ConconiTestAnalyses.objects.select_related('athlete')
            .prefetch_related('conconimeasurement_set')
            .get(pk=event['analysis_pk'])
        )
        measurements = analysis.conconimeasurement_set.order_by('power').values(
            'power', 'heart_rate'
        )

        inputs = {
            'power': [m['power'] for m in measurements],
            'heart_rate': [m['heart_rate'] for m in measurements],
            'weight': float(analysis.athlete.weight),
        }

        report_dir = tempfile.TemporaryDirectory()
        report_filename = '{}.pdf'.format(uuid4().hex)
        report_path = os.path.join(report_dir.name, report_filename)

        results = executor.run('conconi_test', report_path, inputs)

        analysis.result = results['result']

        with open(report_path, 'rb') as fn:
            report_file = File(fn)
            analysis.report = report_file

            analysis.save()

        async_to_sync(self.channel_layer.group_send)(
            event['notify_channel'],
            {
                'type': 'websocket_send',
                'download_url': reverse(
                    'download_report',
                    kwargs={'ref': serialize_model_instance(analysis)},
                ),
            },
        )


class ClientConsumer(JsonWebsocketConsumer):
    """Client websocket consumer."""

    def __init__(self, *args, **kwargs):
        """Initialize client consumer."""
        self.websocket_id = None
        super().__init__(*args, **kwargs)

    def websocket_connect(self, event):
        """Establish websocket."""
        self.websocket_id = self.scope['session'].session_key
        super().websocket_connect(event)

    @property
    def groups(self):
        """Return groups this channel should add itself to."""
        if self.websocket_id is None:
            return []

        return [GROUP_SESSIONS.format(websocket_id=self.websocket_id)]

    def websocket_send(self, event):
        """Send message over websocket."""
        # Strip internal logic.
        event.pop('type')

        self.send_json(event)



"""Utilities for lactolyse analyses."""
import numpy as np


class FittedPolynomial:
    """Polynomial of 3rd degree fitted on the given values."""

    _deriv = None
    _deriv_roots = None
    _second_deriv = None
    _second_deriv_roots = None

    def __init__(self, x_values, y_values):
        """Initialize fitted polynomial."""
        self._x_values = x_values
        self._y_values = y_values

        self.poly = np.poly1d(np.polyfit(x_values, y_values, 3))

        self.error = sum(np.square(self.poly(self._x_values) - self._y_values))

        self.min_x = min(x_values)
        self.max_x = max(x_values)

    @property
    def deriv(self):
        """Return first derivative of the polynomial."""
        if self._deriv is None:
            self._deriv = self.poly.deriv()

        return self._deriv

    @property
    def deriv_roots(self):
        """Return real roots of the first derivative of the polynomial.

        Filter out values that are not between the min and the max value
        of the values on which the polynomial was fitted.
        """
        if self._deriv_roots is None:
            self._deriv_roots = np.roots(self.deriv)

            self._deriv_roots = self._deriv_roots[
                np.logical_and(
                    np.isreal(self._deriv_roots),
                    self._deriv_roots > self.min_x,
                    self._deriv_roots < self.max_x,
                )
            ]

        return self._deriv_roots

    @property
    def second_deriv(self):
        """Return second derivative of the polynomial."""
        if self._second_deriv is None:
            self._second_deriv = self.deriv.deriv()

        return self._second_deriv

    @property
    def second_deriv_roots(self):
        """Return real roots of the second derivative of the polynomial.

        Filter out values that are not between the min and the max value
        of the values on which the polynomial was fitted.
        """
        if self._second_deriv_roots is None:
            self._second_deriv_roots = np.roots(self.second_deriv)

            self._second_deriv_roots = self._second_deriv_roots[
                np.logical_and(
                    np.isreal(self._second_deriv_roots),
                    self._second_deriv_roots > self.min_x,
                    self._second_deriv_roots < self.max_x,
                )
            ]

        return self._second_deriv_roots


def fit_polynomial(x_values, y_values, exclude_outliers=0, return_inputs=False):
    """Fit FittedPolynomial to given values and optionally exclude outliers."""
    best_poly = FittedPolynomial(x_values, y_values)

    for _ in range(exclude_outliers):
        best_x, best_y = x_values, y_values

        for i in range(len(x_values)):
            new_x = x_values[:i] + x_values[i + 1 :]
            new_y = y_values[:i] + y_values[i + 1 :]

            new_poly = FittedPolynomial(new_x, new_y)

            if new_poly.error < best_poly.error:
                best_poly = new_poly
                best_x, best_y = new_x, new_y

        x_values, y_values = best_x, best_y

    if return_inputs:
        return best_poly, x_values, y_values
    else:
        return best_poly



"""Lactolyse utils."""
import json
from base64 import b64decode, b64encode
from importlib import import_module


def serialize_model_instance(instance):
    """Generate a base64 encoded reference to a given Django ORM instance."""
    ref = [instance.__class__.__module__, instance.__class__.__name__, instance.pk]
    dumped_ref = json.dumps(ref).encode('utf-8')
    return b64encode(dumped_ref).decode('utf-8')


def deserialize_model_instance(encoded):
    """Load Django ORM instance from base64 encoded reference."""
    dumped_ref = b64decode(encoded)
    ref = json.loads(dumped_ref.decode('utf-8'))

    model_module = import_module(ref[0])
    model_class = getattr(model_module, ref[1])
    return model_class.objects.get(pk=ref[2])



"""Lactolyse views."""
import os

from django.contrib.auth.decorators import login_required
from django.views.generic import TemplateView

from .conconi_test import ConconiTestView
from .download import ReportDownloadView
from .lactate_threshold import LactateThresholdView

select_analyses_view = login_required(
    TemplateView.as_view(
        template_name=os.path.join('lactolyse', 'select_analyses.html')
    )
)
analyses_success_view = login_required(
    TemplateView.as_view(
        template_name=os.path.join('lactolyse', 'analyses_success.html')
    )
)
lactate_threshold_view = LactateThresholdView.as_view()
conconi_test_view = ConconiTestView.as_view()
report_download_view = ReportDownloadView.as_view()



"""Protocol constants used by Django channels."""

RUN_ANALYSIS_CHANNEL = 'lactolyse.runanalysis'

LACTATE_REPORT_TYPE = 'lactolyse.lactate_report'
CONCONI_REPORT_TYPE = 'lactolyse.conconi_report'

GROUP_SESSIONS = 'lactolyse.session.{websocket_id}'



"""Central place for package metadata."""

from pkg_resources import DistributionNotFound, get_distribution

# NOTE: We use __title__ instead of simply __name__ since the latter would
#       interfere with a global variable __name__ denoting object's name.
__title__ = 'lactolyse'
__summary__ = 'Django package for lactate threshold analyses'
__url__ = 'https://github.com/dblenkus/lactolyse'

try:
    __version__ = get_distribution(__title__).version
except DistributionNotFound:
    # Package is not (yet) installed.
    pass

__author__ = 'Domen Blenkuš'
__email__ = 'domen@blenkus.com'

__license__ = 'Apache License (2.0)'
__copyright__ = '2018, ' + __author__

__all__ = (
    '__title__',
    '__summary__',
    '__url__',
    '__version__',
    '__author__',
    '__email__',
    '__license__',
    '__copyright__',
)



"""Analyses registry."""
import logging

logger = logging.getLogger(__name__)

__all__ = ('registry',)


class AnalysesRegistry:
    _analyses = {}

    def add(self, analysis):
        self._analyses[analysis.name] = analysis

        logger.info("Registered analysis: %s", analysis.name)

    def get(self, analysis_name):
        if analysis_name not in self._analyses:
            raise ValueError(
                "Unknown analysis '{}', select one of the following: {}".format(
                    analysis, ", ".join(self._analyses.keys())
                )
            )

        return self._analyses[analysis_name]


registry = AnalysesRegistry()



"""Lactate threshold analysis."""
import logging

import numpy as np

from .base import BaseAnalysis
from .utils import FittedPolynomial

logger = logging.getLogger(__name__)


class LactateThresholdAnalyses(BaseAnalysis):
    """Lactate threshold analysis."""

    name = 'lactate_threshold'
    template = 'lactate_treshold.tex'

    def _calculate_dmax_context(self, inputs, lac_poly, hr_poly):
        """Calculate context for d-max method."""
        # If polynomial has a local minimum on the interval (of
        # measurments), take it as a minimum value.
        if lac_poly.deriv_roots.size:
            # If there are two roots, we know (based on the shape of the
            # polynomial) that first one is maximum and second one is
            # minimum.
            min_x = max(lac_poly.deriv_roots)

        # If the minimum is not on the interval (or it doesn't exist),
        # we check for the inflection point and take it if it exists.
        elif lac_poly.deriv_roots.size:
            # Second derivation of third degree polynomial have exactly
            # one root (the question is only if it is on the interval).
            min_x = lac_poly.second_deriv_roots[0]

        # If both conditions are false, we can just take the start of
        # the interval, as we know that it is the "most flat" part of
        # the polynomial on the interval.
        else:
            min_x = lac_poly.min_x

        max_x = lac_poly.max_x

        # Find the point where polynomial starts to raise - threshold is
        # 0.5 - and take only real roots (hopefully there is only one).
        roots = np.roots(lac_poly.poly - (lac_poly.poly(min_x) + 0.5))
        roots = roots[np.logical_and(np.isreal(roots), roots > min_x, roots < max_x)]
        start_x = max(roots).real

        # Calculate the vector cross product.
        v_x = np.poly1d(max_x - start_x)
        v_y = np.poly1d(lac_poly.poly(max_x) - lac_poly.poly(start_x))
        u_x = np.poly1d([1, -start_x])
        u_y = lac_poly.poly - lac_poly.poly(start_x)
        cross_z = v_x * u_y - v_y * u_x

        ftp = np.roots(cross_z.deriv())
        ftp = ftp[np.logical_and(ftp > start_x, ftp < max_x)]
        ftp = ftp[0]

        return {
            'power': ftp,
            'start_point': [start_x, lac_poly.poly(start_x)],
            'end_point': [max_x, lac_poly.poly(max_x)],
            'start_hr': hr_poly.poly(start_x),
            'heart_rate': hr_poly.poly(ftp),
            'lactate': lac_poly.poly(ftp),
        }

    def _calculate_cross_context(self, inputs, lac_poly, hr_poly):
        """Calculate context for cross method."""
        if lac_poly.deriv_roots.size:
            start_point = min(lac_poly.deriv_roots)

        else:
            start_point = inputs['power'][0]

        max_x = lac_poly.max_x

        start_line = np.poly1d(
            np.polyfit(
                [start_point, start_point + 5],
                [lac_poly.poly(start_point), lac_poly.poly(start_point + 5)],
                1,
            )
        )
        end_line = np.poly1d(
            np.polyfit(
                [max_x - 5, max_x], [lac_poly.poly(max_x - 5), lac_poly.poly(max_x)], 1
            )
        )

        cross = np.roots(start_line - end_line)
        power = cross[0]

        return {
            'power': power,
            'start_point': [start_point, lac_poly.poly(start_point)],
            'end_point': [inputs['power'][-1], lac_poly.poly(inputs['power'][-1])],
            'cross': [power, start_line(power)],
            'heart_rate': hr_poly.poly(power),
            'lactate': lac_poly.poly(power),
        }

    def _calculate_at_context(self, inputs, threshold, lac_poly, hr_poly):
        """Calculate context for at method."""
        roots = np.roots(lac_poly.poly - threshold)
        roots = roots[np.isreal(roots)]
        roots = filter(
            lambda val: inputs['power'][0] < val < inputs['power'][-1], roots
        )

        power = list(roots)[0].real

        return {
            'power': power,
            'heart_rate': hr_poly.poly(power),
            'lactate': lac_poly.poly(power),
        }

    def render_context(self, inputs):
        """Render the context."""
        for attr in ['power', 'heart_rate', 'lactate']:
            if attr not in inputs:
                raise ValueError("Missing input '{}'.".format(attr))

        lac_poly = FittedPolynomial(inputs['power'], inputs['lactate'])
        hr_poly = FittedPolynomial(inputs['power'], inputs['heart_rate'])

        return {
            'inputs': inputs,
            'lac_poly': lac_poly,
            'dmax': self._calculate_dmax_context(inputs, lac_poly, hr_poly),
            'cross': self._calculate_cross_context(inputs, lac_poly, hr_poly),
            'at2': self._calculate_at_context(inputs, 2, lac_poly, hr_poly),
            'at4': self._calculate_at_context(inputs, 4, lac_poly, hr_poly),
        }

    def get_results(self, context):
        """Return the result of the analysis."""
        return {
            'dmax': context['dmax']['power'],
            'cross': context['cross']['power'],
            'at2': context['at2']['power'],
            'at4': context['at4']['power'],
        }



""".. Ignore pydocstyle D400.

=========
Exectuors
=========

Executor is the central point which takes care to setup the runtime
environment and run all analyses.

Built-in executors are:

.. autoclass:: lactolyse.executors.docker.Executor
    :members:

.. autoclass:: lactolyse.executors.local.Executor
    :members:

"""
import os
import pkgutil
import sys
from importlib import import_module

# Keep Django as a loose dependency.
try:
    from django.core.exceptions import ImproperlyConfigured as error_class
except ImportError:
    error_class = RuntimeError

DEFAULT_EXECUTOR = 'lactolyse.executors.docker'

__all__ = 'executor'


def load_executor(executor_name):
    """Load executor."""
    try:
        return import_module('{}'.format(executor_name))
    except ImportError as err:
        # The executor wasn't found. Display a helpful error message
        # listing all possible (built-in) executors.
        executors_dir = os.path.dirname(__file__)

        try:
            builtin_executors = [
                name for _, name, _ in pkgutil.iter_modules([executors_dir])
            ]
        except EnvironmentError:
            builtin_executors = []
        if executor_name not in [
            'lactolyse.executors.docker.{}'.format(b) for b in builtin_executors
        ]:
            executor_reprs = map(repr, sorted(builtin_executors))
            error_msg = (
                "{} isn't an available executor.\n"
                "Try using 'lactolyse.executors.docker.XXX', where XXX is one of:\n"
                "    {}\n"
                "Error was: {}".format(executor_name, ", ".join(executor_reprs), err)
            )
            raise error_class(error_msg)
        else:
            # If there's some other error, this must be an error in Django
            raise


# Keep Django as a loose dependency.
try:
    from django.conf import settings

    LACTOLYSE_EXECUTOR = getattr(settings, 'LACTOLYSE_EXECUTOR', DEFAULT_EXECUTOR)
except ImportError:
    # Django's settings should be a single source of truth, so environment
    # variable is used only if Django is not installed.
    LACTOLYSE_EXECUTOR = os.environ.get('LACTOLYSE_EXECUTOR', DEFAULT_EXECUTOR)

executor = None
# Initialize executor only in workers.
if len(sys.argv) > 1 and sys.argv[1] == 'runworker':
    executor = load_executor(LACTOLYSE_EXECUTOR).Executor()



"""Conconi test."""
import logging
import math

import numpy as np

from .base import BaseAnalysis
from .utils import fit_polynomial

logger = logging.getLogger(__name__)


class ConcoinAnalyses(BaseAnalysis):

    name = 'conconi_test'
    template = 'conconi_test.tex'

    def _calculate_conconi_context(self, poly, power, heart_rate):
        # FTP value is normally somewhere around 70% of measured values,
        # so we limit the search interval to 60-80% to simplify the
        # prediction.
        lower_limit = power[math.floor(len(power) * 0.6)]
        upper_limit = power[math.ceil(len(power) * 0.8)]

        lower_pwr_range = list(filter(lambda pwr: pwr <= lower_limit, power))
        upper_pwr_range = list(filter(lambda pwr: lower_limit < pwr, power))

        lower_hr_range = heart_rate[: len(lower_pwr_range)]
        upper_hr_range = heart_rate[-len(upper_pwr_range) :]

        min_error = None
        best_lower_line = None
        best_upper_line = None
        while upper_limit > upper_pwr_range[0]:
            lower_line = np.poly1d(np.polyfit(lower_pwr_range, lower_hr_range, 1))
            upper_line = np.poly1d(np.polyfit(upper_pwr_range, upper_hr_range, 1))

            predicted_lower_hr = lower_line(lower_pwr_range)
            predicted_upper_hr = lower_line(upper_pwr_range)

            lower_error = sum(np.square(predicted_lower_hr - lower_hr_range))
            upper_error = sum(np.square(predicted_upper_hr - upper_hr_range))
            error = lower_error + upper_error

            if min_error is None or error < min_error:
                min_error = error
                best_lower_line = lower_line
                best_upper_line = upper_line

            lower_pwr_range.append(upper_pwr_range.pop(0))
            lower_hr_range.append(upper_hr_range.pop(0))

        ftp = np.roots(best_lower_line - best_upper_line)
        ftp = int(ftp[0])

        return {
            'power': ftp,
            'start_point': [power[0], poly.poly(power[0])],
            'end_point': [power[-1], poly.poly(power[-1])],
            'cross': [ftp, best_lower_line(ftp)],
            'heart_rate': poly.poly(ftp),
        }

    def render_context(self, inputs):
        """Render the context."""
        for attr in ['power', 'heart_rate']:
            if attr not in inputs:
                raise ValueError("Missing input '{}'.".format(attr))

        poly, new_pwr, new_hr = fit_polynomial(
            inputs['power'],
            inputs['heart_rate'],
            exclude_outliers=1,
            return_inputs=True,
        )

        return {
            'inputs': inputs,
            'hr_poly': poly,
            'conconi': self._calculate_conconi_context(poly, new_pwr, new_hr),
        }

    def get_results(self, context):
        """Return the result of the analysis."""
        return {'result': context['conconi']['power']}



"""
Django settings for test project.

Generated by 'django-admin startproject' using Django 2.0.2.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '3_%il#)$2=%@$)q793q%q3m@kp$%8%p$r!f(cn%_1h=e55*7^8'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

LOGIN_URL = 'login'
LOGIN_REDIRECT_URL = 'index'


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'channels',
    'material',
    'lactolyse',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ]
        },
    }
]

ASGI_APPLICATION = "tests.routing.application"

CHANNEL_LAYERS = {
    'default': {
        'BACKEND': 'channels_redis.core.RedisChannelLayer',
        'CONFIG': {"hosts": [('localhost', 46379)]},
    }
}


# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.environ.get('LACTOLYSE_POSTGRESQL_NAME', 'lactolyse'),
        'USER': os.environ.get('LACTOLYSE_POSTGRESQL_USER', 'lactolyse'),
        'HOST': os.environ.get('LACTOLYSE_POSTGRESQL_HOST', 'localhost'),
        'PORT': int(os.environ.get('LACTOLYSE_POSTGRESQL_PORT', 45432)),
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'
    },
    {'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator'},
    {'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'},
    {'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'},
]


# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.0/howto/static-files/

STATIC_URL = '/static/'
MEDIA_URL = '/media/'


# Logging.

# Set LACTOLYSE_LOG_FILE environment variable to a file path to enable
# logging debugging messages to to a file.
debug_file_path = os.environ.get('LACTOLYSE_LOG_FILE', os.devnull)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s - %(levelname)s - %(name)s[%(process)s]: %(message)s'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'INFO',
            'formatter': 'standard',
        },
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': debug_file_path,
            'formatter': 'standard',
            'maxBytes': 1024 * 1024 * 10,  # 10 MB
        },
    },
    'loggers': {'': {'handlers': ['file', 'console'], 'level': 'DEBUG'}},
}



"""Lactolyse Django admin."""
from django.contrib import admin

from lactolyse.models import LactateThresholdAnalyses


class LactateThresholdAnalysesAdmin(admin.ModelAdmin):
    """Initial group configuration."""

    model = LactateThresholdAnalyses

    list_display = (
        'get_athlete',
        'created',
        'result_dmax',
        'result_cross',
        'result_at4',
        'result_at2',
    )

    def get_athlete(self, obj):
        """Return athlete's name."""
        return obj.athlete.name

    get_athlete.short_description = 'Athlete'
    get_athlete.admin_order_field = 'athlete__name'


admin.site.register(LactateThresholdAnalyses, LactateThresholdAnalysesAdmin)



""".. Ignore pydocstyle D400.

==================
List Docker images
==================
"""
from django.core.management.base import BaseCommand

from lactolyse.executors.docker import Executor as DockerExecutor


class Command(BaseCommand):
    """List Docker images used by the app."""

    help = "List Docker images used by the app"

    def handle(self, *args, **options):
        """Handle command listdockerimages."""
        images = DockerExecutor.get_docker_images()
        self.stdout.write('\n'.join(images))



"""Lactolyse download views."""
import os

from django.conf import settings
from django.http import HttpResponse, HttpResponseBadRequest
from django.views import View

from lactolyse.models import LactateThresholdAnalyses
from lactolyse.utils import deserialize_model_instance


class ReportDownloadView(View):
    """Download view."""

    def get(self, request, ref):
        """Serve the file specified in Django's session."""
        analysis = deserialize_model_instance(ref)

        if analysis.contributor != request.user and not request.user.is_superuser:
            return HttpResponseBadRequest()

        file_path = os.path.join(settings.MEDIA_ROOT, analysis.report.name)

        with open(file_path, 'rb') as fn:
            response = HttpResponse(fn.read())

        response['Content-Type'] = 'application/pdf'
        response['Content-Length'] = os.path.getsize(file_path)
        response['Content-Disposition'] = 'attachment; filename="Report.pdf"'

        return response



"""Test URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.auth import views as auth_views
from django.urls import include, path


urlpatterns = [
    path(
        'login/',
        auth_views.LoginView.as_view(template_name='lactolyse/login.html'),
        name='login',
    ),
    path('logout/', auth_views.LogoutView.as_view(next_page='index'), name='logout'),
    path('admin/', admin.site.urls),
    path('', include('lactolyse.urls')),
]



"""Lactolyse configuration."""
from django.apps import AppConfig


class LactolyseConfig(AppConfig):
    """Lactolyse AppConfig."""

    name = 'lactolyse'

    def ready(self):
        """Perform application initialization."""
        # Register Analyses.
        from lactolyse.analyses.lactate_threshold import LactateThresholdAnalyses
        from lactolyse.analyses.conconi_test import ConcoinAnalyses

        # Start executor.
        from .executors import executor



import os.path
import setuptools

# Get the long description from the README file
with open('README.rst') as fh:
    long_description = fh.read()

# Get package metadata from '__about__.py' file
about = {}
base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, 'src', 'lactolyse', '__about__.py')) as fh:
    exec(fh.read(), about)

setuptools.setup(
    name=about['__title__'],
    use_scm_version=True,
    description=about['__summary__'],
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author=about['__author__'],
    author_email=about['__email__'],
    url=about['__url__'],
    license=about['__license__'],
    # Exclude tests from built/installed package.
    packages=setuptools.find_packages(
        'src', exclude=['tests', 'tests.*', '*.tests', '*.tests.*']
    ),
    package_dir={'': 'src'},
    package_data={
        'lactolyse': [
            'static/lactolyse/css/*.css',
            'templates/*.html',
            'templates/lactolyse/*.html',
            'templates/lactolyse/latex/*.tex',
        ]
    },
    python_requires='>=3.6, <3.8',
    install_requires=[
        'asgiref~=2.3.2',
        'channels~=2.1.6',
        'channels_redis~=2.3.3',
        'Django~=2.1.5',
        'django-material~=1.4.1',
        'docker~=3.7.0',
        'Jinja2~=2.10',
        'numpy~=1.16.0',
        'psycopg2-binary~=2.7.0',
    ],
    extras_require={
        'package': ['twine', 'wheel'],
        'test': [
            'black',
            'check-manifest',
            'coverage>=4.2',
            'readme_renderer',
            'setuptools_scm',
            'twine',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Other Audience',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='ftp lactate threshold',
)



"""Lactolyse URL configuration."""
from django.urls import path
from django.views.generic import RedirectView

from . import views

urlpatterns = [
    path('', RedirectView.as_view(pattern_name='select_analyses'), name="index"),
    path('tests/', views.select_analyses_view, name='select_analyses'),
    path('tests/lactate/', views.lactate_threshold_view, name='lactate_analyses'),
    path('tests/conconi/', views.conconi_test_view, name='conconi_test'),
    path('tests/success/', views.analyses_success_view, name='analyses_success'),
    path(
        'download/report/<str:ref>/', views.report_download_view, name='download_report'
    ),
]



"""Lactolyse forms."""
from django import forms

from material import Layout, Row

from .models import Athlete, ConconiMeasurement, LactateMeasurement


class AthleteForm(forms.ModelForm):
    """Form for Athlete model."""

    class Meta:
        """AthleteForm Meta options."""

        model = Athlete
        fields = ['name', 'age', 'weight']

    def save(self, contributor, commit=True):
        """Asign the contributor and save the instance."""
        self.instance.contributor = contributor

        return super().save(commit=commit)


class LactateMeasurementForm(forms.ModelForm):
    """Form for LactateMeasurement model."""

    class Meta:
        """LactateMeasurementForm Meta options."""

        model = LactateMeasurement
        fields = ['power', 'heart_rate', 'lactate']

    layout = Layout(Row('power', 'heart_rate', 'lactate'))

    def save(self, analyses, commit=True):
        """Asign the measurement to the analysis and save it."""
        self.instance.analyses = analyses

        return super().save(commit=commit)


class ConconiMeasurementForm(forms.ModelForm):
    """Form for ConconiMeasurement model."""

    class Meta:
        """ConconiMeasurementForm Meta options."""

        model = ConconiMeasurement
        fields = ['power', 'heart_rate']

    layout = Layout(Row('power', 'heart_rate'))

    def save(self, analyses, commit=True):
        """Asign the measurement to the analysis and save it."""
        self.instance.analyses = analyses

        return super().save(commit=commit)



"""Lactolyse models."""

import hashlib
import os

from django.conf import settings
from django.db import models


def _generate_report_name(instance, filename, prefix):
    """Generate unique filename with given prefix."""
    sha256 = hashlib.sha256()
    for chunk in instance.report.chunks():
        sha256.update(chunk)

    extension = os.path.splitext(filename)[1]

    return os.path.join(prefix, sha256.hexdigest() + extension)


def generate_lactate_threshold_name(instance, filename):
    """Generate an unique filename for report based on it's hash."""
    prefix = os.path.join('reports', 'lactate_threshold')
    return _generate_report_name(instance, filename, prefix)


class Athlete(models.Model):
    """Athlete model."""

    contributor = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )

    name = models.CharField(max_length=100)

    age = models.IntegerField()

    weight = models.DecimalField(max_digits=5, decimal_places=2)

    def __str__(self):
        """Format model name."""
        return "name: {}, age: {}".format(self.name, self.age)


class LactateThresholdAnalyses(models.Model):
    """Lactate Threshold Analyses model."""

    contributor = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )

    athlete = models.ForeignKey(Athlete, on_delete=models.CASCADE)

    created = models.DateField(auto_now_add=True)

    report = models.FileField(upload_to=generate_lactate_threshold_name)

    result_dmax = models.IntegerField(null=True)

    result_cross = models.IntegerField(null=True)

    result_at2 = models.IntegerField(null=True)

    result_at4 = models.IntegerField(null=True)

    def __str__(self):
        """Format model name."""
        return "athlete: {}, date: {}".format(self.athlete.name, self.created)


class LactateMeasurement(models.Model):
    """Lactate Measurement model."""

    analyses = models.ForeignKey(LactateThresholdAnalyses, on_delete=models.CASCADE)

    power = models.IntegerField()

    heart_rate = models.IntegerField()

    lactate = models.DecimalField(max_digits=4, decimal_places=2)

    def __str__(self):
        """Format model name."""
        return "name: {}, power: {}, heart rate: {}, lactate: {}".format(
            self.analyses.athlete.name, self.power, self.heart_rate, self.lactate
        )


class ConconiTestAnalyses(models.Model):
    """Lactate Threshold Analyses model."""

    contributor = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )

    athlete = models.ForeignKey(Athlete, on_delete=models.CASCADE)

    created = models.DateField(auto_now_add=True)

    report = models.FileField(upload_to=generate_lactate_threshold_name)

    result = models.IntegerField(null=True)

    def __str__(self):
        """Format model name."""
        return "athlete: {}, date: {}".format(self.athlete.name, self.created)


class ConconiMeasurement(models.Model):
    """Lactate Measurement model."""

    analyses = models.ForeignKey(ConconiTestAnalyses, on_delete=models.CASCADE)

    power = models.IntegerField()

    heart_rate = models.IntegerField()

    def __str__(self):
        """Format model name."""
        return "name: {}, power: {}, heart rate: {}".format(
            self.analyses.athlete.name, self.power, self.heart_rate
        )



import logging
import shutil
import subprocess
import tempfile
import time

from lactolyse.registry import registry

logger = logging.getLogger(__name__)


class Executor:
    """Executor using local environment."""

    def run(self, analysis, output_path, inputs=None):
        """Run the analysis."""
        if inputs is None:
            inputs = {}

        if not isinstance(inputs, dict):
            raise ValueError("Attribute 'inputs' must be of type 'dict' or 'None'.")

        with tempfile.TemporaryDirectory() as runtime_dir:
            analysis = registry.get(analysis)(runtime_dir=runtime_dir)

            start_time = time.time()

            logger.debug("Running run_render_context method.")
            analysis.run_render_context(inputs)

            logger.debug("Running render_template method.")
            analysis.render_template()

            logger.debug("Compiling report.")
            result = subprocess.run(
                analysis.run_command, cwd=runtime_dir, shell=True, capture_output=True
            )

            logger.info("Report finished in {:.2f}s.".format(time.time() - start_time))

            if result.returncode != 0:
                message = "Something went wrong while rendering report."
                logger.error(message)
                print(result.stdout)
                raise RuntimeError(message)

            shutil.copy(analysis.get_pdf_file(), output_path)

        return analysis.run_get_results()



import atexit
import logging
import os
import shutil
import tempfile
import time

import docker

from lactolyse.registry import registry

logger = logging.getLogger(__name__)

DOCKER_IMAGE = 'domenblenkus/lactolyse:latest'
DOCKER_START_COMMAND = "/bin/sh -c 'sleep infinity'"
DOCKER_MOUNT_POINT = '/mnt'

docker_client = docker.from_env()


def remove_docker_container(container_id):
    """Return the function which removes the Docker container with thhe given id when called."""

    def _remove_container():
        """Remove the Docker container."""
        try:
            container = docker_client.containers.get(container_id)
        except docker.errors.NotFound:
            # Looks like container has already been removed.
            return

        # container.stop()
        container.remove(force=True)

    return _remove_container


class Executor:
    """Executor using Docker container for the environment.

    This executor starts a Docker container at the initiazlization and
    executes all commands inside it.
    """

    def __init__(self):
        self._runtime_root = tempfile.TemporaryDirectory()
        logger.info("Runtime root: %s", self._runtime_root.name)

        self._container = None

        self._create_container(ignore_errors=True)

    def _create_container(self, ignore_errors=False):
        """Create the container."""
        runtime_mount = docker.types.Mount(
            DOCKER_MOUNT_POINT, self._runtime_root.name, 'bind'
        )

        logger.debug("Creating container.")
        try:
            self._container = docker_client.containers.run(
                DOCKER_IMAGE, DOCKER_START_COMMAND, mounts=[runtime_mount], detach=True
            )
        except docker.errors.ImageNotFound:
            if ignore_errors:
                logger.warning("Container image not found.")
                return

            logger.exception("Container image not found.")
            raise

        logger.info("Container created: %s", self._container.id)

        # Remove the container on exit.
        atexit.register(remove_docker_container(self._container.id))

    def _check_container(self):
        """Check that container is up and running.

        Reload the container status and start it if it is not running
        already. If the container doesn't exist anymore, create a new
        one.
        """
        if self._container is None:
            self._create_container()

        try:
            self._container.reload()
        except docker.errors.NotFound:
            logger.warning("Container cannot be found, creating a new one.")
            self._create_container()

            return

        if self._container.status != 'running':
            logger.warning("Container was not running, starting it.")
            self._container.start()

    @classmethod
    def get_docker_images(self):
        """Return list of Docker images used by executor."""
        return [DOCKER_IMAGE]

    def run(self, analysis, output_path, inputs=None):
        """Run the analysis."""
        if inputs is None:
            inputs = {}

        if not isinstance(inputs, dict):
            raise ValueError("Attribute 'inputs' must be of type 'dict' or 'None'.")

        self._check_container()

        with tempfile.TemporaryDirectory(dir=self._runtime_root.name) as runtime_dir:
            analysis = registry.get(analysis)(runtime_dir=runtime_dir)

            start_time = time.time()

            logger.debug("Running run_render_context method.")
            analysis.run_render_context(inputs)

            logger.debug("Running render_template method.")
            analysis.render_template()

            logger.debug("Compiling report.")
            result = self._container.exec_run(
                "/bin/sh -c 'cd {}; {}'".format(
                    os.path.join(DOCKER_MOUNT_POINT, os.path.basename(runtime_dir)),
                    analysis.run_command,
                )
            )

            logger.info("Report finished in {:.2f}s.".format(time.time() - start_time))

            if result.exit_code != 0:
                message = "Something went wrong while rendering report."
                logger.error(message)
                print(result.output)
                raise RuntimeError(message)

            shutil.copy(analysis.get_pdf_file(), output_path)

        return analysis.run_get_results()
